import logging, base64, io
import requests
import shutil
import mimetypes
from lxml import etree
import json
import pathlib

DEFAULT_TRANSPORT = 'Symphony'
DEFAULT_POLL_RETRY = 30
DEFAULT_TIMEOUT = 60


class AriaService:
    def __init__(self, base_url, session_id, **kwargs):
        self._base_url = base_url
        self._token = None
        self.delta_token = ''
        self._paging_token = None
        self._session_id = session_id
        self._timeout_default = 30 if 'timeout' not in kwargs else int(kwargs['timeout'])
        self._headers = {'content-type':'application/json'}

    def _process_aria_message(self, data, attribute=None):
        if 'detail' in data:
            detail = data['detail']
            if detail is not None and attribute in detail:
                return detail.get(attribute)
            else:
                return detail
        else:
            raise AriaError('Unrecognised server response', data)

    def authenticate_app(self, key):
        token = self._authenticate('appKey', key)
        self._headers['Authorization'] = 'Bearer {}'.format(token)

    def authenticate(self, key):
        token = self._authenticate('userKey', key)
        self._headers['Authorization'] = 'Bearer {}'.format(token)

    def _authenticate(self, keyType, key):
        data = {
            "session": self._session_id,
            "credentials": [{'transport': 'Symphony', keyType: key.decode('utf-8')}]
        }
        res = requests.post(
            self._base_url+'/auth',
            json=data,
            headers={'content-type': 'application/json'}
        )
        res.raise_for_status()
        return self._process_aria_message(res.json(), 'token')

    def get_messages(self, delta_token=None):
        if delta_token is not None:
            self.delta_token = delta_token
        res = requests.get(
            self._base_url+'/message',
            headers=self._headers,
            timeout=self._timeout_default,
            params={'deltaToken': self.delta_token, 'isSubscription': True}
        )
        res.raise_for_status()
        data = res.json()
        if 'deltaToken' in data:
            self.delta_token = data.get('deltaToken')
        retval = self._process_aria_message(data)
        if retval is None:
            return []
        else:
            return retval

    def search_messages(self, data, timeout=None, limit=1000):
        timeout = self._timeout_default if timeout is None else timeout

        res = requests.post(
            self._base_url+'/message/search',
            json=data,
            headers=self._headers,
            timeout=timeout,
            params={'limit': limit}
        )
        res.raise_for_status()
        return self._process_aria_message(res.json())

    def send_message(self, channel, raw_message, **kwargs):
        message = {"channel": channel, "payload": {}}
        data = None if 'data' not in kwargs else kwargs['data']
        if isinstance(data, str):
            data = json.loads(data)

        message = AriaService.set_text(message, raw_message, data)
        if 'attachment' in kwargs and kwargs['attachment'] is not None:
            if isinstance(kwargs['attachment'], tuple):
                filename, fdata = kwargs['attachment']
                message['attachments'] = []
                message['attachments'].append(
                    {'name': filename,
                     'data': base64.b64encode(fdata).decode(),
                     'content-type': mimetypes.guess_type(filename)}
                )
        if 'on_behalf' in kwargs and kwargs['on_behalf'] is not None:
            message['onBehalf'] = kwargs['on_behalf']
        return requests.post('{}/message'.format(self._base_url), json=message, headers=self._headers)

    def delete_message(self, id):
        payload = {"id": id}
        res = requests.delete('{}/message'.format(self._base_url), params=payload, headers=self._headers)
        res.raise_for_status()
        return res

    def get_user(self, user_id, id_type='id', admin=False):
        params = {id_type: user_id}
        if admin:
            params['admin'] = True
        res = requests.get(
            self._base_url+'/user', params=params,
            headers=self._headers,
            timeout=self._timeout_default
        )
        res.raise_for_status()
        return self._process_aria_message(res.json())

    def who_am_i(self):
        res = requests.get(
            self._base_url+'/user/session',
            headers=self._headers,
            timeout=self._timeout_default
        )
        res.raise_for_status()
        return self._process_aria_message(res.json(), 'email')

    def get_connections(self):
        res = requests.get(
            self._base_url+'/connection',
            headers=self._headers,
            timeout=self._timeout_default
        )
        res.raise_for_status()
        return self._process_aria_message(res.json())

    def add_connection(self, user, accept=None):
        data = {"user": user}
        if accept:
            data['accept'] = "true"
        elif not accept:
            data['accept'] = "false"

        res = requests.post(
            self._base_url+'/connection',
            headers=self._headers,
            json=data,
            timeout=self._timeout_default
        )
        res.raise_for_status()
        return self._process_aria_message(res.json())

    def get_room_details(self, room_id, id_type='id', limit=1):
        params = { id_type: room_id, 'limit': limit }
        res = requests.get(
            '{}/room'.format(self._base_url),
            params=params,
            headers=self._headers,
            timeout=self._timeout_default
        )
        if res.status_code == requests.codes.not_found:
            raise AriaError('RoomId {0} not found'.format(room_id), res)
        else:
            res.raise_for_status()

        return self._process_aria_message(res.json())

    def create_room(self, room_name, room_description):
        res = requests.post(
            '{}/room'.format(self._base_url),
            json={"name": room_name, "description": room_description},
            headers=self._headers,
            timeout=self._timeout_default
        )
        res.raise_for_status()
        return self._process_aria_message(res.json())

    '''Display all members of a room'''
    def get_room_members(self, room_name, **kwargs):
        params = {}
        if 'on_behalf' in kwargs and kwargs['on_behalf'] is not None:
            params['on_behalf'] = kwargs['on_behalf']

        res = requests.get(
            '{}/room/{}/member'.format(self._base_url, room_name),
            headers=self._headers,
            params=params,
            timeout=self._timeout_default
        )
        if res.status_code == requests.codes.not_found:
            raise AriaError('Room {0} not found'.format(room_name), res)
        else:
            res.raise_for_status()
        return self._process_aria_message(res.json())

    '''Add a member or list of members to a room'''
    def add_room_members(self, room_name, member_email, **kwargs):
        users = []
        if not isinstance(member_email, list):
            users.append(member_email)
        else:
            for user in member_email:
                users.append(user)

        data = {"email": users}
        if 'on_behalf' in kwargs and kwargs['on_behalf'] is not None:
            data['on_behalf'] = kwargs['on_behalf']

        res = requests.post(
            '{}/room/{}/member'.format(self._base_url, room_name),
            json=data,
            headers=self._headers,
            timeout=self._timeout_default
        )
        if res.status_code == requests.codes.not_found:
            raise AriaError('Room {0} not found'.format(room_name), res)
        else:
            res.raise_for_status()
        return self._process_aria_message(res.json())

    '''Remove a member or list of members from a room'''
    def remove_room_members(self, room_name, member_email, **kwargs):
        users = []
        if not isinstance(member_email, list):
            users.append(member_email)
        else:
            for user in member_email:
                users.append(user)

        data = {"email": users}
        if 'on_behalf' in kwargs and kwargs['on_behalf'] is not None:
            data['on_behalf'] = kwargs['on_behalf']

        res = requests.delete(
            '{}/room/{}/member'.format(self._base_url, room_name),
            params=data,
            headers=self._headers,
            timeout=self._timeout_default
        )
        if res.status_code == requests.codes.not_found:
            raise AriaError('Room {0} not found'.format(room_name), res)
        else:
            res.raise_for_status()
        return self._process_aria_message(res.json())

    def download_file(self, url, local_filename=None):
        if not local_filename:
            local_filename = url.split('/')[-1]
        r = requests.get(url, headers=self._headers, stream=True)
        r.raise_for_status()
        body = r.json()
        logging.debug('File: {}'.format(str(body)))
        with open(local_filename, 'wb') as f:
            bfile = io.BytesIO(base64.b64decode(body['message']))
            shutil.copyfileobj(bfile, f)
        return local_filename

    #    def send_stream_as_attachment(self, channel, attach, stream, disableEncoding=True, passthrough=True):

    @staticmethod
    def set_text(message, text, data=None):
        if isinstance(text, SymphonyMessageML):
            message['payload']['message'] = text.tostring()
        else:
            message['payload']['message'] = text
        if data is not None:
            message['payload']['data'] = json.dumps(data)
        return message

    @staticmethod
    def read_cert(filename):
        return AriaService.base64_encode_file(filename)

    @staticmethod
    def base64_encode_file(filename):
        if not pathlib.Path(filename).exists():
            raise AriaError('File {} does not exist'.format(filename), None)
        with open(filename,'rb') as f:
            return base64.b64encode(f.read())


class AriaError(Exception):
    def __init__(self, text, res):
        self.text = text
        self.response = res


class SymphonyMessageML:
    def __init__(self):
        self._root = etree.Element('messageML')

    def tostring(self):
        return etree.tostring(self._root).decode('utf-8')

    def root(self):
        return self._root

    def text(self, text):
        self._root[len(self._root)-1].text = text
        return self

    def tail(self, text):
        self._root[len(self._root)-1].tail = text
        return self

    def bold(self, text=None):
        _elem = etree.SubElement(self._root, 'b')
        if text is not None:
            _elem.text = text
        return self

    def span(self, text=None):
        _elem = etree.SubElement(self._root, 'span')
        if text is not None:
            _elem.text = text
        return self

    def paragraph(self, text=None):
        _elem = etree.SubElement(self._root, 'p')
        if text is not None:
            _elem.text = text
        return self

    def anchor(self, url=None):
        _elem = etree.SubElement(self._root, 'a')
        if url is not None:
            _elem.attrib['href'] = url
        return self

    def tbreak(self, text=None):
        _elem = etree.SubElement(self._root, 'br')
        if text is not None:
            _elem.text = text
        return self

    def hrule(self, text=None):
        _elem = etree.SubElement(self._root, 'hr')
        if text is not None:
            _elem.text = text
        return self

    def italic(self, text=None):
        _elem = etree.SubElement(self._root, 'i')
        if text is not None:
            _elem.text = text
        return self

    def __str__(self):
        return self.tostring()


__all__ = ['AriaService', 'SymphonyMessageML', 'AriaError']



#!/usr/bin/env python3
# coding: utf8

import argparse
import pathlib
import re
import importlib
import sys
from time import sleep
from requests.exceptions import HTTPError, RequestException, ConnectionError

import ariaclient.api
from ariaclient import AriaError, AriaService, SymphonyMessageML
from sys import modules as MODULES
import json
import logging

from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

CONFIG_REQUIRED=frozenset(['url', 'session', 'cert'])
MESSAGE_REQUIRED=frozenset(['text', 'channel', 'from'])

logging.basicConfig(format="%(asctime)-15s %(levelname)s-%(message)s",
                    level=logging.DEBUG, datefmt='%Y-%m-%d %H:%M:%S')


class AriaBot:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.config = kwargs['config']
        if 'aria' not in self.config:
            raise AttributeError('Missing aria section in config')
        self.env = self.config['aria']
        self._populate_commands()
        self._populate_entity_handlers()
        self.session = None
        self.aria = None
        self._delta_token = ''
        self.read_token_file = pathlib.Path(self.env.get('read_token_file', 'read_token_file.dat'))
        if not self.read_token_file.exists():
            logging.warning('Token file {0} doesnt exist, creating new file'.format(self.read_token_file.name))
            self.read_token_file.write_text('')

    def _populate_commands(self):
        if 'commands' not in self.config:
            logging.warning("No commands configured")
        else:
            for _cmd in self.config['commands']:
                flags = 0
                if 'flags' in _cmd:
                    flags = eval(_cmd['flags'])
                _cmd['pattern'] = re.compile(_cmd['pattern'], flags)
                # push the full config to all commands
                _cmd['config'] = self.config
                _module = None
                try:
                    if 'module' in _cmd:
                        _module = importlib.import_module(_cmd['module'])
                    else:
                        _module = MODULES[__name__]
                    _class = getattr(_module, _cmd['class'])
                    _cmd['bot'] = self
                    _class(**_cmd)
                except AttributeError as ae:
                    logging.warning('{} - ignoring command'.format(ae.args[0], _cmd['class']))

    def _populate_entity_handlers(self):
        if 'entities' not in self.config:
            logging.warning("No entities configured")
        else:
            for _ent in self.config['entities']:
                _module = None
                try:
                    if 'module' in _ent:
                        _module = importlib.import_module(_ent['module'])
                    else:
                        _module = MODULES[__name__]
                    _class = getattr(_module, _ent['class'])
                    _ent['bot'] = self
                    _class(**_ent)
                except AttributeError as ae:
                    logging.warning('{} - ignoring entity'.format(ae.args[0], _ent['class']))

    def _process_message(self, channel, msg, msg_id, attachments, msg_text):
        processed = False
        start = 0
        item = 1
        if not isinstance(attachments, list):
            attachments = None
        while start < len(msg_text):
            cmd, cmd_match = AriaBotCommand.match_command(msg_text, start)
            if not cmd_match:
                break
            processed = True
            cmd.do(match=cmd_match, message=msg, msg_id=msg_id, index=item, attachments=attachments, channel=channel)
            start = cmd_match.span()[1] + 1
            item = item + 1
        return processed

    def _process_entities(self, msg, data):
        processed = False
        item = 1
        # Check for "io.fincloud.biz.objectlist"
        obj_type = data.get('type', None)
        if obj_type == "io.fincloud.biz.objectlist":
            obj_list = data.get('list', None)
            obj_header = data.get('header', None)
        else:
            obj_list = [data]
            obj_header = None

        # Loop through all the entities in the message
        for entity in obj_list:
            # Check if there is a 'type' attribute on the entity and process.
            if isinstance(entity, dict) and 'type' in entity:
                handler = AriaBotEntityHandler.match_entity(entity['type'])
                if not handler:
                    logging.debug('Unmatched entity {}'.format(entity['type']))
                    continue
                handler.handle(entity=entity, message=msg, index=item, header=obj_header)
                processed = True
            item = item + 1
        return processed

    def start(self):
        logging.info('Starting ARIA Bot version[{0}] - python version [{1}'.format(ariaclient.__version__, sys.version))
        authenticated = False
        poll_retry = ariaclient.api.DEFAULT_POLL_RETRY
        timeout = ariaclient.api.DEFAULT_TIMEOUT
        system = self.config.get('system', None)
        if system and 'poll_retry' in system:
            poll_retry = int(system['poll_retry'])
        if system and 'timeout' in system:
            timeout = int(system['timeout'])
        if not CONFIG_REQUIRED.issubset(list(self.env.keys())):
            raise AriaBotException('Missing required attribute from: '+str(CONFIG_REQUIRED))

        cer = None
        try:
            cer = AriaService.read_cert(self.env['cert'])
            self._delta_token = self.read_token_file.read_text()
            logging.info('Resuming aria stream [{0}]'.format(self._delta_token))
        except AriaError as e:
            logging.error(e)
            sys.exit(1)

        self.aria = AriaService(
            self.env['url'],
            self.env['session'],
            timeout=timeout
        )
        self.session = None
        # cross_pod = ('cross_pod' in self.env and not self.env['cross_pod'])
        while True:
            # reset these on every iteration
            channel = None
            msg_text = ''
            msg = None
            try:
                if not authenticated:
                    self.aria.authenticate(cer)
                    self.session = self.aria.who_am_i()
                    authenticated = True

                for msg in self.aria.get_messages(delta_token=self._delta_token):
                    # Check the cross-pod flag and ignore message if not valid.
                    # if (not cross_pod) and msg['header']['crossPod']:
                    #    logging.debug('Ignore cross-pod message from [{}]'.format(msg['from']))
                    #    continue

                    logging.debug('Msg: {}'.format(str(msg)))
                    # Ignore everything except actual messages, or anonymous
                    if msg['type'] != 'MESSAGE_SENT' or 'email' not in msg['sender']:
                        continue
                    from_email = msg['sender']['email']
                    # if this message is from the bot skip to next message.
                    if from_email == self.session:
                        continue

                    msg_text = msg['text']
                    channel = from_email
                    if 'channel' in msg:
                        channel = msg['channel']['name']
                    payload = msg['payload']
                    processed_text = self._process_message(
                        channel,
                        payload.get('message', ''),
                        payload.get('messageId', ''),
                        msg['attachments'], msg_text
                    )
                    processed_entities = False
                    if 'data' in payload:
                        data = json.loads(payload['data'])
                        processed_entities = self._process_entities(msg, data)

                    if (not processed_text) and (not processed_entities):
                        logging.debug('Not matched {}'.format(msg_text))

            except (ValueError, AriaBotException) as ex:
                logging.error(ex.args[0])
                if msg is not None:
                    err_msg = SymphonyMessageML()
                    err_msg.root().text = '🤖 [{}]'.format(ex.args[0])
                    self.aria.send_message(channel, err_msg)

            except (KeyboardInterrupt, SystemExit) as exiting:
                logging.error('Exiting: {}'.format(exiting))
                break

            except HTTPError as hte:
                status = hte.response.status_code
                logging.error('Error: {} {}'.format(status, hte.response.text))
                if status == 401:
                    authenticated = False
                    sleep(1)
                elif status == 404:
                    # Datafeed not found - renewing
                    logging.warning(hte.response.text+' renewing datafeed')
                    # Blank the delta token for the next read, causing a refresh
                    self.aria.delta_token = ''
                    self._delta_token = ''
                elif status == 502:
                    logging.error('Gateway error: {} retry in {} seconds'.format(hte.response.status_code, poll_retry))
                    sleep(poll_retry)
                else:
                    logging.error('Error[{}] req[{}] res[{}] '.format(status, hte.request.body, hte.response.text))
                    if channel is not None:
                        err_msg = SymphonyMessageML()
                        err_msg.root().text = '🤖 [{}]'.format(msg_text)
                        self.aria.send_message(channel, err_msg)
                    sleep(poll_retry)

            except ConnectionError as ce:
                logging.error('Connection Error: {}'.format(ce))
                sleep(poll_retry)

            except RequestException as reqE:
                logging.error('Request Exception: {}'.format(reqE))
                sleep(poll_retry)
            # if aria client has a token value from a new session, then write this to disk
            if self.aria.delta_token != self._delta_token:
                self._delta_token = self.aria.delta_token
                self.read_token_file.write_text(self._delta_token)


class AriaBotCommand:

    _commands = []

    def __init__(self, **kwargs):
        self.bot = kwargs['bot']
        self.pattern = kwargs['pattern']
        self.name = kwargs['name']
        self.helptext = kwargs.get('helptext', None)
        self.config = kwargs['config']
        AriaBotCommand._commands.append(self)

    @staticmethod
    def match_command(msg_text, start):
        for cmd in AriaBotCommand._commands:
            match = cmd.pattern.match(msg_text, start)
            if match:
                return cmd, match
        return None, None

    def do(self, **kwargs):
        pass

    def simple_help(self):
        help_msg = SymphonyMessageML().bold('🤖: Here\'s a list of commands I understand')
        help_msg.hrule()
        for cmd in self._commands:
            if cmd.helptext is not None:
                help_msg = help_msg.bold(cmd.name).tail(' {} '.format(cmd.helptext)).tbreak()
            else:
                help_msg = help_msg.bold(cmd.name).tbreak()
        return help_msg


class AriaBotEntityHandler:

    _handlers = {}

    def __init__(self, **kwargs):
        self.bot = kwargs['bot']
        self.entity = kwargs['entity']
        self.name = kwargs['name']
        AriaBotEntityHandler._handlers[self.entity] = self

    @staticmethod
    def match_entity(entity):
        if entity in AriaBotEntityHandler._handlers.keys():
            return AriaBotEntityHandler._handlers[entity]
        else:
            return None

    def handle(self, **kwargs):
        pass


class AriaBotException(Exception):
    pass


__all__ = ['AriaBotException', 'AriaBot', 'AriaBotEntityHandler', 'AriaBotCommand']

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='ARIA Bot Framework')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('configfile', help='configuration file')
    parser.add_argument('environment', help='environment, referenced in config section')
    args = parser.parse_args()

    allConfig = load(open(args.configfile, 'r'), Loader=Loader)

    if args.verbose:
        logging.basicConfig(format="%(asctime)-15s %(levelname)s-%(message)s",
                            level=logging.DEBUG, datefmt='%Y-%m-%d %H:%M:%S')
    else:
        logging.basicConfig(format="%(asctime)-15s %(message)s", level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
        logging.getLogger("requests").setLevel(logging.WARNING)

    _bot = AriaBot(config=allConfig, environment=args.environment)
    _bot.start()
